from collections import namedtuple

PrefixFunctions = namedtuple("PrefixFunctions", ['prefix', 'hasit', 'filterit', 'makeit', ])

def prefix_funcs(prefix):

    def hasit(x):
        return x.startswith(prefix)

    def filterit(lst):
        return [ x for x in lst if hasit(x) ]

    def makeit(x):
        if hasit(x):
            return x
        else:
            return prefix + x

    def removeit(x):
        if hasit(x):
            return x.replace(prefix, '', 1)

    return PrefixFunctions(prefix, hasit, filterit, makeit)

PrefixesFunctions = namedtuple("PrefixesFunctions", ['prefixes', 'hasem', 'filterem', 'funcs'])

def prefixes_funcs(prefixes):

    def hasone(x):
        return any((x.startswith(p) for p in prefixes))

    def filterem(lst):
        return [ x for x in lst if hasone(x) ]

    return PrefixesFunctions(prefixes, hasone, filterem, [prefix_funcs(p) for p in prefixes])




import logging

from flib.env import args

from fabric import colors
from pprint import pformat

class ColorFormatter(logging.Formatter):

    def format(self, record):
        bold = False
        if record.levelno >= logging.ERROR:
            color = colors.red
        elif record.levelno >= logging.WARNING:
            color = colors.yellow
        elif record.levelno >= logging.INFO:
            if args.debug:
                bold = True
            color = colors.cyan
        else:
            if args.debug:
                color = colors.cyan
            else:
                color = colors.white

        if isinstance(record.msg, dict):
            record.msg = pformat(record.msg)
        msg = super(ColorFormatter, self).format(record)
        if msg.endswith('\n'):
            return color(msg[:-1], bold)
        else:
            return color(msg, bold)

class UpToLogLevel(logging.Filter):

    def filter(self, record):
        if record.levelno < self.upTo:
            return True
        else:
            return False

def getConsoleHandlers(args, config):
    '''creates a handler for console output'''
    import sys
    ok_handler = logging.StreamHandler(sys.stdout)
    if args.debug:
        ok_handler.setLevel(logging.DEBUG)
    else:
        ok_handler.setLevel(logging.INFO)

    nok_filter = UpToLogLevel()
    nok_filter.upTo = logging.WARNING
    ok_handler.addFilter(nok_filter)

    errhandler = logging.StreamHandler(sys.stderr)
    errhandler.setLevel(nok_filter.upTo)

    if args.nofmt or args.list:
        formatter = logging.Formatter('%(message)s')
    else:
        formatter = ColorFormatter('%(message)s')
    ok_handler.setFormatter(formatter)
    errhandler.setFormatter(formatter)
    return ok_handler, errhandler


def getFileHandler(filename, args, config):
    '''creates a file handler'''
    handler = logging.FileHandler(filename)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    return handler

def configure_logger(name, args=None, config=None):
    '''create a logger configured from args and config'''

    if args is None:
        from flib.env import args

    if config is None:
        from flib.env import config

    logger = logging.getLogger(name)
    logger.setLevel(1)
    ok_handler, errhandler = getConsoleHandlers(args, config)
    logger.addHandler(ok_handler)
    logger.addHandler(errhandler)
    for fname in args.output:
        logger.addHandler(getFileHandler(fname, args, config))
    return logger


def list_commands(cmd_dct):
    logger = configure_logger('list_commands')
    logger.info("Available commands:")
    for cmd in sorted(cmd_dct):
        func = cmd_dct[cmd]
        doc = func.__doc__
        if doc is not None and not args.verbose:
            doc = doc.split('\n')[0]
        if args.nofmt:
            msg = "    %-16s%s" % (cmd, doc)
        else:
            msg = "    %-25s%s" % (colors.green(cmd), colors.cyan(doc))
        logger.info(msg)

from flib import lst2cmd
def log_cmd(command, *cmd_args):
    if 0 < args.verbose < 3:
        log = logging.getLogger('command')
        if args.nofmt:
            log.info(' $ %s %s' % (command, lst2cmd(cmd_args)))
        else:
            cmd = colors.yellow(' $ %s %s' % (command, lst2cmd(cmd_args)), True)
            log.info(cmd)

def log_cwd_cmd(cwd, command, *cmd_args):
    if 0 < args.verbose < 3:
        log = logging.getLogger('command')
        if args.nofmt:
            log.info('%s $ %s %s' % (cwd, command, lst2cmd(cmd_args)))
        else:
            cmd = colors.yellow('%s %s' % (command, lst2cmd(cmd_args)), True)
            log.info('%s $ %s' % (cwd, cmd))

def log_putget(pg, src, dest):
    if 0 < args.verbose < 3:
        log = logging.getLogger('command')
        if args.nofmt:
            log.info(' %s %s %s' % (pg, src, dest))
        else:
            cmd = colors.yellow(' %s %s %s' % (pg, src, dest), True)
            log.info(cmd)

def log_result(result):
    if args.verbose == 2:
        log = logging.getLogger('results')
        if result.stdout:
            log.info(result.stdout)
        if result.stderr:
            log.warn(result.stderr)
    elif args.verbose >= 3:
        log = logging.getLogger('results')
        if args.nofmt:
            cmdline = result.cmdline
            exit_code = result.exit_code
        else:
            if not result.exit_code:
                cmdline = colors.white(result.cmdline, True)
                exit_color = colors.green
            else:
                cmdline = colors.red(result.cmdline, True)
                exit_color = colors.red
        log.info('ran: %s' % cmdline)
        if args.verbose > 4:
            log.info('in: %s' % exit_color(repr(result.cwd)))
        else:
            log.info('in: %s' % exit_color(result.cwd))
        if result.stdout or args.verbose > 3:
            if args.verbose == 3:
                log.info('stdout:\n%s' % result.stdout)
            else:
                log.info('stdout:\n%r' % result.stdout)
        if result.stderr or args.verbose > 3:
            if args.verbose == 3:
                log.warn('stderr:\n%s' % result.stderr)
            else:
                log.warn('stderr:\n%r' % result.stderr)
        log.info('return: %s' % exit_color(result.exit_code))
    return result



from flib.cmd import expose
from flib.flow import prefix_funcs
from flib.env import args, config
from flib.output import configure_logger
from flib import configured
from flib import abort

log = configure_logger('simple_flow')

repo = configured.path_obj(git=True)

log.debug(config)

master = config.flow.master
develop = config.flow.develop

ft = prefix_funcs(config.flow.feature)
rl = prefix_funcs(config.flow.release)

@expose(docargs=True)
def feature(ftargs):
    '''Manage feature branches.

    Usage:
        feature NAME
        feature [ -n | --new ] NAME
        feature [ -u | --update ] NAME
        feature [ -c | --continued ] NAME
        feature [ -f | --finish ] NAME

    Options:
        -n, --new         Create a new feature.
        -u, --update      Update a feature.
        -c, --continued   A feature was continued.
        -f, --finish      Finish a feature.
    '''
    log.debug(ftargs)
    name = ftargs.NAME

    if ftargs.new:
        feature = ft.makeit(name)
        return new_feature(feature)

    feature = repo.get_branch(name)

    if ftargs['--update']:
        return update_feature(feature)

    elif ftargs.continued:
        return continued_feature(feature)

    elif ftargs.finish:
        result = update_feature(feature)
        return finish_feature(feature, result)

    else:
        repo.git('checkout', feature)


def new_feature(feature):
    log.info('Will create %r in %s.' % (feature, repo))
    repo.git('checkout', master)
    repo.git('checkout', '-b', feature)
    log.info('Enjoy %s.' % feature)

def update_feature(feature):
    repo.git('checkout', feature)
    log.info('Update %s.' % feature)
    repo.git('merge', master)

def finish_feature(feature, update_result):
    log.info('Merge into %s.' % master)
    repo.git('checkout', master)
    repo.git('merge', feature)
    log.info('Clean up %s.' % feature)
    repo.git('branch', '-d', feature)

def continued_feature(feature):
    log.info('Sry, not yet impl. ;-p')


@expose
def remaster():
    '''Update all branches that are based on master.'''
    branches = repo.local_branches()
    for b in ft.filterit(branches):
        update_feature(b)



from flib.env import args, config
from flib.local import LocalHost
from flib.remote import RemoteHost
from flib.repo import GitRepository
from flib.repo import Directory

def env_host():
    if args.host:
        return args.host
    elif 'host' in config:
        return config.host
    elif 'repo' in config and 'host' in config.repo:
        return config.repo.host
    else:
        return None

_cache = None
def host_obj():
    global _cache
    if _cache is not None:
        return _cache
    cfg = env_host()
    if cfg:
        _cache = RemoteHost(cfg)
    else:
        _cache = LocalHost()
    return _cache


def env_path():
    if args.path != '.':
        return args.path
    elif 'repo' in config and 'location' in config.repo:
            return config.repo.location
    else:
        return args.path

def path_obj(git=True):
    cls = GitRepository if git else Directory
    return cls(host_obj(), env_path())



import pytest

from flib.local import LocalHost
from flib.remote import RemoteHost

fixtures = {
    'hostclass': [LocalHost, RemoteHost],
    'hostname': ['localhost', 'tuxcode.org'],
    'host': [LocalHost(), RemoteHost('localhost')],
}

def pytest_generate_tests(metafunc):
    for name, values in fixtures.iteritems():
        if name in metafunc.fixturenames:
            metafunc.parametrize(name, values)

@pytest.fixture
def localhost():
    return LocalHost()

def test_local_ls(localhost):
    ls = localhost.sh("ls")
    assert ls.exit_code == 0

def test_local_echo(localhost):
    echo = localhost.sh("echo", "hello world")
    echoecho = localhost.sh("echo", "hello", "world")
    assert echo != echoecho
    assert echo.exit_code == 0
    assert echo.stdout == "hello world\n"
    assert echo.stderr == ""
    assert echoecho.exit_code == 0
    assert echoecho.stdout == "hello world\n"
    assert echoecho.stderr == ""
    assert echo.cmdline == "/bin/bash -l -c \"echo 'hello world'\""
    assert echoecho.cmdline == "/bin/bash -l -c 'echo hello world'"

#def test_ls(host):
    #assert host.sh("ls")



from flib.cmd import expose
from flib.env import args, config
from flib.output import configure_logger
from flib import abort
from fabric.contrib.console import confirm

log = configure_logger('default_cmds')

@expose
def clean():
    '''Clean pyc and other python temp files.'''
    import os
    to_clean = []
    for loc, dirs, files in os.walk(os.curdir):
        to_clean.extend([os.path.join(loc, f) for f in files if f.endswith('.pyc')])
    log.info('About to delete:\n%s' % '\n'.join(to_clean))
    do_delete = confirm('Ok to delete these files/dirs?')
    log.debug(do_delete)
    if do_delete:
        map(os.remove, to_clean)
        log.info('done.')

@expose
def get_branch(nam):
    '''get_branch test'''
    from flib.repo import GitRepository
    from flib.local import LocalHost
    repo = GitRepository(LocalHost(), '.')
    log.info(repo.get_branch(nam))



from inspect import isfunction
from flib.output import configure_logger
from flib.env import args, config
from flib import abort, lst2cmd
from flib import abort

log = configure_logger('objmappers')

from collections import namedtuple
InexistingShellResult = namedtuple('InexistingShellResult', ['cmdline', 'cwd', 'stdout', 'stderr', 'exit_code'])

class Directory(object):
    '''Base class for a directory on some host.'''

    def __init__(self, host=None, path=None, not_there=args.dirs):
        self.host = host
        self.path = path
        self.exists = self.check_exists(path)
        self.not_there = not_there
        if 'import_deferred' in config:
            config.import_deferred.append((self.cwd, [path], {}))
        else:
            self.cwd(path, not_there=not_there)

    #@property
    #def host(self, host):
        #if isinstance(host, str):
            #if host 

    def check_exists(self, path=None):
        if path is None:
            path = self.path
        try:
            return self.host.sh('test', '-d', path).exit_code == 0
        except Exception, ex:
            log.error(ex)
            return False

    def create(self, path=None):
        if path is None:
            path = self.path
        log.info('Creating %s.' % self)
        result = self.host.sh('mkdir', '-p', self.path)
        if result.exit_code == 0:
            if path == self.path:
                self.exists = True
                self.sh = self.host.bake(cwd=path)
            return True
        else:
            if path == self.path:
                self.exists = False
                self.sh = self.not_there_sh
            return False

    def not_there_sh(self, command, *args):
        cmd = '%s %s' % (command, lst2cmd(args))
        if self.not_there == 'ignore':
            return InexistingShellResult(cmd, self.path, '', '', 0)
        else:
            msg = 'Not exisitng %s ignores command: %s' % (self, cmd)
            log.warn(msg)
            return InexistingShellResult(cmd, self.path, '', '', 1)

    def __repr__(self):
        return "%s(%s, %r)" % (self.__class__.__name__, self.host, self.path)

    def cwd(self, path, not_there=None):
        self.path = path
        self.exists = self.check_exists(path)

        if not_there is None:
            not_there = self.not_there

        if self.exists or not_there == 'ignore':
            self.sh = self.host.bake(cwd=path)
            return True
        else:
            self.sh = self.not_there_sh

        if not_there == 'create':
            return self.create()
        elif not_there == 'abort':
            abort(log, 'Error: %r is not a directory!' % (self,))
        else:
            assert not_there == 'warn'
            log.warn('Warning: %r is not a directory!' % (self,))
            return False


class GitRepository(Directory):

    def __init__(self, host, path, not_there=args.gits):
        if not_there == 'init':
            super(GitRepository, self).__init__(host, path, not_there='create')
        else:
            super(GitRepository, self).__init__(host, path, not_there=not_there)
        self.not_there = not_there

    def not_there_git(self, *args):
        cmd = 'git %s' % lst2cmd(args)
        if self.not_there == 'ignore':
            return InexistingShellResult(cmd, self.path, '', '', 0)
        else:
            msg = 'Uninitialized %s ignores command: %s' % (self, cmd)
            log.warn(msg)
            return InexistingShellResult(cmd, self.path, '', '', 1)

    def _git(self, *args):
        git = self.host.bake('git', cwd=self.path)
        return git('-c', 'color.ui=false', *args)

    def cwd(self, path, not_there=None):
        if not_there is None:
            not_there = self.not_there

        if not_there == 'init':
            if self.exists or self.check_exists():
                super(GitRepository, self).cwd(path)
            else:
                super(GitRepository, self).cwd(path, not_there='create')
        else:
            super(GitRepository, self).cwd(path, not_there=not_there)

        if not self.exists:
            self.git = self.not_there_git
        else:
            self.git = self._git

        if not self.git('rev-parse', '--is-inside-work-tree').exit_code == 0:
            if not_there == 'ignore':
                pass
            elif not_there == 'abort':
                abort(log, 'Error: %r is not a repository!' % (self,))
            elif not_there == 'init':
                log.info('Initializing %s.' % self)
                self.git('init', self.path)
            else:
                assert not_there in ('warn', 'create')
                log.warn('Warning: %s is not a repository.' % (self,))

    def _branches(self):
        result = []
        current = None
        for line in self.git("branch", "-l").stdout.split('\n'):
            if line.startswith('* '):
                log.debug(line)
                line = line[2:]
                current = line
            if line:
                result.append(Branch(line.strip()))
        log.debug(result)
        return result, current

    def local_branches(self):
        return self._branches()[0]

    def current_branch(self):
        return self._branches()[1]

    def get_branches(self, filter_thing):
        '''Return all branches that match the filter criteria.'''
        if hasattr(filter_thing, 'hasit'):
            return filter(filter_thing.hasit, self.local_branches())
        elif hasattr(filter_thing, 'hasone'):
            return filter(filter_thing.hasone, self.local_branches())
        elif isfunction(filter_thing):
            return filter(filter_thing, self.local_branches())

    def get_branch(self, part, on_many='abort'):
        '''convenience method for commandline input'''
        possible = [b for b in self.local_branches() if part in b]
        if len(possible) == 1:
            return possible.pop()
        elif on_many == 'abort':
            log.error('Given branch part does not identify exactly one branch: %s' % possible)
        else:
            return possible

    def bake_branch(self, name):
        return Branch(self, name)

class Branch(str):

    def __init__(self, name, repo=None):
        str.__init__(self, name)
        self.name = name
        self.repo = repo



from flib.cmd import expose
from flib.flow import prefix_funcs
from flib.env import args, config
from flib.output import configure_logger

log = configure_logger('test_flow')

@expose("alias")
def aliased():
    # an undocumented command
    log.info("this is aliased")
    log.debug("You're in debug, and i know it.")

@expose
def aliased():
    """A documented command"""
    log.info("this is weird")
    log.debug(config['module'])


@expose(docargs=True)
def tryme(*myargs):
    '''usage: ft-try [-cl]

        -c, --cool      be cool as heck
        -l, --lame      be lame as hell
    '''
    log.info(repr(myargs))


from flib.local import LocalHost
from flib.repo import Directory, GitRepository
from flib.configured import host_obj

@expose
def local():
    """Some local commands"""
    host = LocalHost()

    d = host.bake_dir('/tmp/foo')
    d.sh("ls")
    #d.git('-c', 'color.ui=false', 'status')
    d.sh('ls | grep boot')
    d.sh('ls | grep boot')

@expose("config")
def cfg():
    """Use stuff from flowcfg/"""
    log.info(config)

@expose
def putget():
    """Some local puts and gets"""
    host = LocalHost()

    host.put('ft.py', '/tmp')
    host.get('/tmp/ft.py', 'tmp')

    ok_put = host.put('ft.py', '/tmp')
    nok_put = host.put('ft.py', '/')

    ok_get = host.get('/tmp/ft.py', 'tmp')
    nok_get = host.get('/tmp/xxx', 'tmp')


from flib.remote import RemoteHost

@expose
def remote():
    """Some remote commands"""
    rhost = RemoteHost('localhost')

    rd = rhost.bake_dir('/home/k', git=True)
    rd.sh("ls")
    rd.git('-c', 'color.ui=false', 'status').stdout

    d = rhost.bake_dir('/', git=True)
    d.sh("ls")
    d.git('-c', 'color.ui=false', 'status').stderr
    d.sh('ls | grep boot').stdout
    d.sh('ls | grep boot').exit_code

@expose
def rputget():
    """Some remote puts and gets"""
    rhost = RemoteHost('localhost')

    ok_put = rhost.put('ft.py', '/tmp')
    nok_put = rhost.put('ft.py', '/')

    ok_get = rhost.get('/tmp/ft.py', 'tmp')
    nok_get = rhost.get('/tmp/xxx', 'tmp')



from flib.host import Host
from flib import lst2cmd
from fabric import api

from functools import wraps
from flib.env import args
from flib import ShellResult

def fab2res(r):
    return ShellResult(r.real_command, r.cwd, r.stdout, r.stderr, r.return_code)

def fabputget2res(pg, s, d, r):
    return ShellResult('%s(%s, %s)' % (pg, s, d), d, tuple(r), r.succeeded, int(not r.succeeded))

def quietly(func):
    context = api.warn_only if args.debug else api.quiet
    @wraps(func)
    def wrapped(*args, **kwd):
        with api.settings(context(), use_ssh_config=True):
            return func(*args, **kwd)
    return wrapped

class RemoteHost(Host):

    def __init__(self, name, user=None, port=None, key_file=None):
        if user is None and not "@" in name:
            name = "%s@%s" % (api.env['user'], name)
        elif user is not None:
            name = "%s@%s" % (user, name)
        self.user, self.name = name.split("@")
        self.login = name
        self.port = None
        self.key_file = key_file

    @quietly
    def _sh(self, cwd, *args):
        '''emulate sh.command(*args)'''
        def run():
            with api.settings(api.cd(cwd), port=self.port, key_filename=self.key_file):
                result = api.run(lst2cmd(args), pty=False)
                result.cwd = cwd
                return result
        result = fab2res(api.execute(run, hosts=[self.login])[self.login])
        assert result.exit_code == 0
        return result

    @quietly
    def handle_command(self, *args):
        '''emulate sh.command(*args)'''
        def run():
            with api.settings(port=self.port, key_filename=self.key_file):
                result = api.run(lst2cmd(args), pty=False)
                result.cwd = ''
                return result
        result = fab2res(api.execute(run, hosts=[self.login])[self.login])
        assert result.exit_code == 0
        return result

    #@quietly
    #def run(self, command):
        #def run():
            #return api.run(command, pty=False)
        #result = api.execute(run, hosts=[self.login])
        #return fab2res(result[self.login])

    #@quietly
    #def sudo(self, command):
        #def run():
            #return api.sudo(command, pty=False)
        #result = api.execute(run, hosts=[self.login])
        #return fab2res(result[self.login])

    @quietly
    def _put(self, source, dest):
        def run():
            return api.put(source, dest)
        result = api.execute(run, hosts=[self.login])
        return fabputget2res('put', source, dest, result[self.login])

    @quietly
    def _get(self, source, dest):
        def run():
            return api.get(source, dest)
        result = api.execute(run, hosts=[self.login])
        return fabputget2res('get', source, dest, result[self.login])



from flib import lst2cmd
from flib.env import args as global_args
from flib.output import configure_logger
from flib.output import log_cmd, log_cwd_cmd, log_putget, log_result
log = configure_logger('BaseHost')
configure_logger('command')
configure_logger('results')

from flib.repo import GitRepository
from flib.repo import Directory
from flib import ShellResult

class Host(object):
    '''Base class for hosts of all sorts.'''

    name = None
    user = None
    login = None

    def __str__(self):
        if not hasattr(self, 'login'):
            return "%s()" % (self.__class__.__name__)
        else:
            return "%s(%r)" % (self.__class__.__name__, self.login)

    def sh(self, command, *args):
        log_cmd(command, *args)
        if global_args.notreally:
            cmd = ' $ %s %s' % (command, lst2cmd(args))
            return log_result(ShellResult(cmd, None, '', '', 0))
        else:
            return log_result(self.handle_command(command, *args))

    def handle_command(self, command, *args):
        self._handle_cwd_command(self, '.', command, *args)

    def _handle_cwd_command(self, cwd, command, *args):
        log_cwd_cmd(cwd, command, *args)
        if global_args.notreally:
            cmd = '%s $ %s %s' % (cwd, command, lst2cmd(args))
            return log_result(ShellResult(cmd, cwd, '', '', 0))
        else:
            result = self._sh(cwd, command, *args)
            log_result(result)
            return result

    def _sh(self, cwd, command, *args):
        raise NotImplementedError('Base host class does not implement _sh')


    def bake(self, command=None, cwd=None):
        if command is None:
            def baked(*args):
                log.debug('cwd bakery: %r %r' % (cwd, args))
                return self._handle_cwd_command(cwd, *args)
        else:
            if cwd is None:
                def baked(*args):
                    log.debug('cmd bakery: %r %r' % (command, args))
                    return self.sh(command, *args)
            else:
                def baked(*args):
                    log.debug('cwdcmd bakery: %r %r %r' % (cwd, command, args))
                    return self._handle_cwd_command(cwd, command, *args)
        return baked


    def bake_dir(self, path, git=False):
        if git:
            return GitRepository(self, path)
        else:
            return Directory(self, path)


    def put(self, *args):
        log_putget('>', *args)
        if global_args.notreally:
            cmd = 'put(%r, %r)' % args
            return log_result(ShellResult(cmd, tuple(args), '', '', 0))
        else:
            return log_result(self._put(*args))

    def _put(self, src, dest):
        raise NotImplementedError('Base host class does not implement _put')

    def get(self, *args):
        log_putget('<', *args)
        if global_args.notreally:
            cmd = 'get(%r, %r)' % args
            return log_result(ShellResult(cmd, tuple(args), '', '', 0))
        else:
            return log_result(self._put(*args))

    def _get(self, src, dest):
        raise NotImplementedError('Base host class does not implement _get')
